/*  Sql2ClassConsole.java
 *	Copyright (c) 	2001, Dennis Groenendijk
 *					2002-2004, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: 21 mei 2001
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
* [When]			[Who]							[What]
* 06-06-2001   dennis@brains2b.nl		Added catalog as option, user not longer required
* 28-08-2002   dennis@brains2b.nl		made initialization more public 
* 08-12-2002		"			"					closed connection when no longer needed.
* 08-03-2003 	"			"  				Improved displaying info an logging
* 04-11-2003		"			"					Implemented Main class and TypeWriter project
* 														Removed unnecessary accessors
* 22-03-2004	dennis@brains2b.nl		Added options from parameters
*/
package org.brains2b.sql.toclass;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Properties;
import java.util.Vector;

import org.brains2b.console.ArgumentException;
import org.brains2b.console.ArgumentMap;
import org.brains2b.console.Main;
import org.brains2b.log.Logger;
import org.brains2b.sql.Connector;
import org.brains2b.sql.meta.SqlMetaData;
import org.brains2b.sql.meta.SqlTableModel;

/** Console Application for running Sql2Class
 *  <p>Sql2Class is a function which transforms a Database table into a java Class
 *  and writes a control class to interact with the database.
 *  <p>Sql2Class takes a number of parameters:
 *  <table><tr><th>parameter</th<th>short form</th><th>required</th><th>description</th></tr>
 *  <tr><td>driver</td><td>d</td><td>yes</td><td>qualified class name of the JDBC driver to use</td>
 *  <tr><td>url</td><td>r</td><td>yes</td><td>JDBC url to connect to the datasoource</td>
 *  <tr><td>user</td><td>u</td><td>no</td><td>username for the datasource, required only if JDBC connection requires it</td>
 *  <tr><td>password</td><td>p</td><td>no</td><td>password for the datasource, required only if JDBC connection requires it</td>
 *  <tr><td>table</td><td>t</td><td>yes</td><td>The table or view you want to generate</td>
 *  <tr><td>location</td><td>l</td><td>no</td><td>the path where to write the newly generated classes to. The package name
 * 	will be appended to the end of this location. <i>If no location is defined the directory where Sql2Class was started is used</td>
 *  <tr><td>package</td><td>g</td><td>no</td><td>The package for this class. This will be both set in de class files as used
 * to append to the path as directory to create the classes in.</td>
 *  <tr><td>readonly</td><td>w</td><td>no</td><td>Define that the control should be generated to as read-only</td></tr>
 *  <tr><td>optimistic</td><td>o</td><td>no</td><td>Use the OptimisticLocking and control code to identify the 
 *  data class as using optimistic locking. No locking mechanisme is providided if this is not set</td></tr>
 *   <tr><td>forcerowid</td><td>f</td><td>no</td><td>Use a provided unique row identification instead of primary keys,
 *  or unique indexes to generate the data control. If there are no primary keys or unique indexes this option will be 
 *  used by default.</td></tr>
 *  <p>Sql2Class can take the necessary parameters from a property file <code>sql2class.properties</code> or from the commandline.
 * The commandline overwrites parameters set in de sql2class.properties file. Use the parameter names to define the properties.
 *  <p>parameters passed to the commandline should be prefixed with <code>--</code> if you use the full name or <code>-</code>
 * if you use the short name.
 * <p>If the files you are generating  are the same as the files at the supplied location and package no new files
 * will be generated. Look at the <code>SqlClassWriter.isFileSame(File)</code> for the conditions.
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.30 [Nov 4, 2003]
 * @see SqlClassWriter#isFileSame(File)
 */
public class Sql2ClassConsole extends Main {

	public static final String VERSIONID="$0.30$";

    /**
     * Constructor
     * <p>only sets the argument map to use for this application
     *
     */
    public Sql2ClassConsole() {
    	init();
    }

	/**
	 * main
	 * @param args String[], the commandline arguments
	 */
    public static void main(String[] args) {
        try {
            Sql2ClassConsole sql = new Sql2ClassConsole();
            sql.initialize(args);
            sql.makeClass();
        } catch (Exception ex) {
           Logger.printException(ex);
        }
    }

	/** initializes the commandline arguments for Sql2Class
	 * @see ArgumentMap
	 * @see Main
	 */ 
	private void init() {
		ArgumentMap map=new ArgumentMap();
		map.addOption("table",'t',true);
		map.addArgumentTo("table","tablename",true);
		map.addOption("driver",'d',true);
		map.addArgumentTo("driver","name",true);
		map.addOption("url",'r',true);
		map.addArgumentTo("url","name",true);
		map.addOption("user",'u',false);
		map.addArgumentTo("user","name",true);
		map.addOption("password",'p',false);
		map.addArgumentTo("password","word",true);
		map.addOption("catalog",'c',false);
		map.addArgumentTo("catalog","name",true);
		map.addOption("location",'l',false);
		map.addArgumentTo("location","dir",true);
		map.addOption("package",'g',false);
		map.addArgumentTo("package","name",true);
		map.addOption("readonly",'w',false);
		map.addOption("forcerowid",'f',false);
		map.addOption("optimistic",'o',false);
		setArgumentMap(map);
	}

	/** initialize Sql2Class with the values from properties
	 * @param prop Properties, the properties containing the arguments for Sql2Class
	 * @throws ArgumentException, thrown if a property from the file does not 
	 */
	protected void initialize(Properties prop) throws ArgumentException {
		Iterator it=prop.keySet().iterator();
		String[] arguments = new String[prop.size()*2];
		int i=0;
		while (it.hasNext()) {
			String key=(String) it.next();
			arguments[i++] = "--" + key;
			arguments[i++] = prop.getProperty(key);
		}
		makeArgumentList(arguments);
	}

	/**
	 * initialize Sql2Class with the values from the arguments and properties
	 *  by first reading the properties and then setting the arguments if provided
	 * @param args String[], the command line arguments
	 * @throws Exception
	 */
    private void initialize(String[] args) throws Exception {
        Properties p=readProperties();
        String[] arguments = new String[args.length+p.size()*2];
        Iterator it=p.keySet().iterator();
        int i=0;
        while (it.hasNext()) {
        	String key=(String) it.next();
        	arguments[i++] = "--" + key;
        	arguments[i++] = p.getProperty(key);
        }
        System.arraycopy(args,0,arguments,p.size()*2,args.length);
        makeArgumentList(arguments);
    }

	/**
	 * the method which actually creates the classes
	 * <p>This method checks first if the files exists and are the same
	 * as the classes which are to be created. If new classes are to be created
	 * the table is looked up in the database and two seperate threads are started
	 * to create the data class and the data control class
	 * @throws SQLException, thrown if the JDBC connection
	 * cannot complete normally
	 */
    protected void makeClass() throws SQLException {
    	//make sure user name is in capitals
    	setArgument("user",getArgument("user").toUpperCase());
    	
        File fl_location = null;
        if (getArgument("location")!=null) {
            fl_location = new File(getArgument("location"));
        }
        Sql2File  s2f = new Sql2File(fl_location, getArgument("package"));
        boolean s2fExists = s2f.fileExists(getArgument("table"));
        Sql2DataC s2d = new Sql2DataC( fl_location, getArgument("package"));
		boolean s2dExists = s2d.fileExists(getArgument("table"));
		if (s2fExists && s2dExists) {
			return;
		}
		
        Connector con = null;
        SqlMetaData meta = null;
        SqlTableModel tm = null;
        //Vector vc = null;
        try {
            // connect to the database
            con = new Connector(
            	getArgument("driver"),
            	getArgument("url"),
            	getArgument("user"),
            	getArgument("password"));
            meta = new SqlMetaData(con.getConnection());
            // get table information
            Logger.println("finding Table: " + getArgument("table"), Logger.DEBUG);
            Vector vt = meta.getTables(getArgument("catalog"), getArgument("user"), getArgument("table"));
            if (vt.size() != 1) {
                throw new SQLException(
                    "No table found, or more then one with name " +getArgument("table"));
            } else {
                tm = (SqlTableModel) vt.get(0);
            }
           
            Logger.println("finding Columns: " + getArgument("table"), Logger.DEBUG);
            
            tm.setColumns(meta.getColumns(getArgument("catalog"), getArgument("user"), getArgument("table"), null));
           
            	Logger.println("finding PrimaryKeys: " + getArgument("table"), Logger.DEBUG);
            	tm = meta.getPrimaryKey(tm, getArgument("catalog"), getArgument("user"), getArgument("table"),"true".equals(getArgument("forcerowid")));
          
        } catch (SQLException se) {
            throw se;
        } finally {
        	con.close();
        }
		if (!s2fExists) {
	        s2f.setTableModel(tm);
			s2f.setOptimisticLock("true".equals(getArgument("optimistic")));
    	    Thread t = new Thread(s2f);
        	t.run();
		}
		if (!s2dExists) {
			s2d.setTableModel(tm);
			s2d.setReadOnly("true".equals(getArgument("readonly")));
    	    Thread t = new Thread(s2d);
        	t.run();
		}

    }

    /**
     * read properties from the default file
     * @return Properties, a properties object containing all the values set through the file
     * @throws Exception
     */
    private Properties readProperties() throws Exception {
		Properties pp = new Properties();
        try {
            File fl = new File(mPropFile);
            pp.load(new FileInputStream(fl));
        } catch (IOException ie) {
            //do nothing
        } catch (NullPointerException npex) {
            //do nothing
        } 
		return pp;
        
    }

    private String mPropFile = "sql2class.properties";
  
}
