/*  Sql2DataC.java
 *	Copyright (c) 2002, Brains2B.org
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
 * [When]			[Who]					[What]
 * 08/29/2002	dennis@brains2b.nl 	Filled in selectOnCondition
 * 09/06/2002	dennis@brains2b.nl	Corrected problems in thet SQL statements
 * 11/25/2002		"				"				Uses getPreparedStatement from condition 
 * 02-02-2003		"				"				do not write complete functions for read only (views) and tables without a primary key (excl. ROWID)
 * 08-03-2003 		"				" 				Improved displaying info an logging
 * 23-03-2004 	dennis@brains2b.nl		Added javadoc, return right result and message on error
 * 														Implemented readOnly and optimistic lock as options
 * TODO ROWID is only valid for Oracle!
 */
package org.brains2b.sql.toclass;

import java.io.File;
import java.lang.reflect.Modifier;
import java.util.Date;
import java.util.Vector;

import org.brains2b.log.Logger;
import org.brains2b.sql.meta.SqlColumnModel;
import org.brains2b.sql.meta.SqlTableModel;
import org.brains2b.sql.meta.Types2Java;
import org.brains2b.typewriter.ClassTemplate;
import org.brains2b.typewriter.CommentTemplate;
import org.brains2b.typewriter.ContentTemplate;
import org.brains2b.typewriter.IfTemplate;
import org.brains2b.typewriter.MethodTemplate;
import org.brains2b.util.StringHelper;

/** This class generates a DataControl class in line with the DataC specifications
 * to control insert,delete,update and select statements for a given data class.
 * <p>This class is called from <code>Sql2ClassConsole</code> and some other implementations
 * which supply the necessary information. If you use this class just to generate files look at 
 * <code>Sql2ClassConsole</code> first. 
 * <p>The naming of the new class file will be equal to the table name with initial
 *  caps and followed by DC. example: <i>EMPLOYEES will become EmployeesDC</i>. Underscores will be treated
 * as an instance to initialize caps again. example: <i>EMP_SALARIES will become EmpSalariesDC</i>
 * <p>The class is written to the given location and packageName. The location should not include the
 * package directory itself. If the package directory does not exist it will be created. If the location
 * does not exist you will get a <code>FileNotFoundException</code>
 * <p>The option <a href="#setOptimisticLock><code>setOptimisticLock</code></a> sets the code in the data class to 
 * implement optimistic locking. If you leave this to false, no locking mechanism is provided 
 * <p>The option <a href="#setReadOnly><code>setReadOnly</code></a> generates a data class which will throw
 * an <code>Exception</code> and performs no other function if you try to update,insert or delete a record through this data control, making this
 * data control basically a read only data control.
 * <p>If you generate a data control as a view the insert,update and delete will be generated throwing only an 
 * <code>Exception</code> while updating views is not supported.
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.40 [23-03-2004]
 * @see org.brains2b.sql.toclass.Sql2ClassConsole
 * @see org.brains2b.sql.meta.SqlTableModel
 */
public class Sql2DataC extends SqlClassWriter {

	private String m_name;
	private String m_objName;
	private String m_instance;
	private boolean m_readOnly;
	private boolean m_optimisticLock;
	
	private ClassTemplate classTemp;
	
	/**
	 * Constructor
	 * <p>This constructor does not set the TableModel, this is however required
	 * for this class to be able to generate the datacontrol. You should set the tablemodel
	 * by calling <code>setTableModel(org.brains2b.sql.meta.SqlTableModel)</code>
	 * @param location File, set the location to write the file to, this location should exist before run is called
	 * on this class.
	 * @param packageName String, the packageName for the class, names should be separated by dots. The packagename
	 * is both used to set the package name of the class as to create or use the directory relative to location to write the file to.
	 */
	public Sql2DataC( File location, String packageName) {
		super(location, packageName);
	}

	/**
	 * Constructor
	 * @param location File, set the location to write the file to, this location should exist before run is called
	 * on this class.
	 * @param packageName String, the packageName for the class, names should be separated by dots. The packagename
	 * is both used to set the package name of the class as to create or use the directory relative to location to write the file to.
	 * @param tm org.brains2b.sql.meta.SqlTableModel, the table model to base this dataclass on
	 */
	public Sql2DataC( File location, String packageName, SqlTableModel tmodel) {
		super(location, packageName);
		setTableModel(tmodel);
	}

	/**
	 * run does the actual work in this class.
	 * <p>Output is written to the <code>org.brains2b.log.Logger</code>. Dependant
	 * on the setting there the code will show up on <code>System.out</code>. This is
	 * the case if you do nothing
	 * @see java.lang.Runnable#run()
	 * @see org.brains2b.task.DefaultTask#setFinished(boolean)
	 * @see org.brains2b.log.Logger
	 */
	public void run() {
		if (m_tableModel != null) {
			final boolean readOnly = isReadOnly() || m_tableModel.getTableType().equals("VIEW");
			try {
				m_name=getClassName(m_tableModel.getTableName());
				m_instance = varName(m_objName);
				
				classTemp=new ClassTemplate(m_packageName,
					m_name,"org.brains2b.sql.oracle.OracleDC",null);
				classTemp.addImport("java.util.Vector");
				classTemp.addImport("java.sql.ResultSet");
				classTemp.addImport("java.sql.Connection");
				
				writeHeaderComment();
				progress();
				writeConstructor();
				progress();
				writeGetObject();
				progress();
				if (readOnly) {
					writeUpdateforReadOnly();
				} else {
					writeUpdateOnPk();
				}
				progress();
				if (readOnly) {
					writeInsertforReadOnly();
				} else {
					writeInsert();
				}
				progress();
				if (readOnly) {
					writeDeleteforReadOnly();
				} else {
					writeDeleteOnPk();
				}
				progress();
				writeSelectOnParent();
				progress();
				if (getPrimaryKey(m_tableModel) != null) {
					writeSelectOnPk();
				} else {
					writeSelectforNonPk();
				}
				progress();
				writeSelectAll();
				progress();
				progress();
				writeSelectFromCondition();
				progress();
				classTemp.writeClass(m_location.getAbsolutePath());
				Logger.println("datacontrol class saved as " + m_objName + ".java");
				setResult(Sql2DataC.RESULT_OK);
			} catch (Exception ex) {
				if (!"NOTHING TO DO".equals(ex.getMessage())) {
					Logger.printException(ex);
					setResult(Sql2DataC.RESULT_ERROR);
					addMessage(ex.getMessage());
				}
			} finally {
					setFinished(true);
			}
		}

	}
	
	/**
	 * get the file which has been created. 
	 * <p>The result of this file is only valid after <code>run()</code>has finished
	 * @see org.brains2b.sql.toclass.SqlClassWriter#getResultfile()
	 * @see #run()
	 */
	public File getResultfile() {
		return classTemp.getCompleteFile();
	}
	
	/**
	 * write the comment for the start of the class file
	 */
	private void writeHeaderComment() {
		CommentTemplate comment=classTemp.createComment();
		comment.addText(Sql2ClassConsole.VERSIONID);
		comment.addText("Class " + m_name);
		comment.addText("Generated from "+ m_tableModel.getTableType().toLowerCase() + ": " + m_tableModel.getTableName());
		comment.addText("Schema: " + m_tableModel.getTableSchema());
		//XXX get database name
		comment.addText("Database: ");
		comment.addText(" ");
		comment.addText("@since: " + new Date().toString());
		
	}
	
	/**
	 * write the default constructor for a Data Control class
	 * taking both a <code>java.sql.Connection</code> and <code>Object<code> 
	 */
	private void writeConstructor() {
				
		ContentTemplate constructor=
			classTemp.createConstructor(new String[][] {
				{"Connection","Object"}, {"con","o"}
			}); 
		constructor.addText("super(con,o); ");
		
	}

	/**
	 * write the general getObject statement which fills the data object with the values
	 * supplied by a resultset.
	 * <p>sets the appropriate checksum value if optimistic locking is enabled
	 * @see #setOptimisticLock(boolean)
	 */
	private void writeGetObject() {
		final String oName = StringHelper.objectName(m_tableModel.getTableName());
		final String vName = m_instance;
		MethodTemplate get=classTemp.createMethod(Modifier.PROTECTED,
			"getObject","Object",new String[][] {{"ResultSet"},{"rs"}});
		get.addException("Exception");
		
		get.addText(oName + " " + vName + " = new " + oName +"(");
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			if (cm.getColumnName().equals("ROWID")) {
				get.addText("\trs.getString(" + (i + 1) + ")"
				+ (i == m_tableModel.getColumns().size() - 1?");":","));
			} else {
				get.addText(
					"\trs.get"
						+ StringHelper.initCaps(
							Types2Java.getJavaDataType(
								cm.getDataType(),
								cm.getColumnSize(),
								cm.getDecimalDigits()),
							' ')
						+ "("
						+ cm.getOrdinalPosition()
						+ ")"
						+ (i == m_tableModel.getColumns().size() - 1?");":","));
			}
		}
		if (isOptimisticLock()) get.addText(vName + ".setChecksum(" + vName + ".hashCode());");
		get.addText("return " + vName + ";");

	}

	/**
	 * write the update for read only statement containing only a throw Exception
	 *
	 */
	private void writeUpdateforReadOnly() {
		MethodTemplate t=classTemp.createMethod("updateOnPk","int",null);
		t.addException("Exception");
		writeReadOnlyException(t);
		
	}

	/**
	 * write the insert for read only statement containing only a throw Exception
	 *
	 */
	private void writeInsertforReadOnly() {
		MethodTemplate t=classTemp.createMethod("insert","int",null);
		t.addException("Exception");
		writeReadOnlyException(t);
		
	}
	
	/**
	 * write the delete for read only statement containing only a throw Exception
	 *
	 */
	private void writeDeleteforReadOnly() {
		MethodTemplate t=classTemp.createMethod("deleteOnPk","int",null);
		t.addException("Exception");
		writeReadOnlyException(t);
		
	}

	/**
	 * write the selectOnPk statements for statements containing no
	 * primary key, pseudo key or ROWID only adding a throws clause 
	 *
	 */
	private void writeSelectforNonPk() {
		MethodTemplate t=classTemp.createMethod(Modifier.PUBLIC,"selectOnPk","Object",
		new String[][] {{"Object"},{"o"}});
		t.addException("Exception");
		
		t.addText(
			"throw new SQLException(\"Function is not available, This table has no primary key\");");
		
	}
	
	/**
	 * General message used in the forReadOnly methods to throw an SQLException
	 * saying the function is not available, because the table is read only
	 * @param t ContentTemplate, the template the exception code should be added to
	 */
	private void writeReadOnlyException(ContentTemplate t) {
		t.addImport("java.sql.SQLException");
		t.addText(
			"throw new SQLException(\"Function is not available, This table is read only\");");
	}

	/**
	 * writes a normal updateOnPk method, possibly with optimistic locking code
	 * @see #setOptimisticLock(boolean)
	 */
	private void writeUpdateOnPk() {
		SqlColumnModel[] pk = getPrimaryKey(m_tableModel);
		if (pk==null) {
			throw new RuntimeException("Table model has no primary key, pseudo primary key or ROWID. It is not possible to generate the DataControl");
		}
		MethodTemplate t=classTemp.createMethod("updateOnPk","int",null);
		t.addException("Exception");
		
		
		t.addText(m_objName+ " "+ m_instance+ " = (" + m_objName	+ ") mObj;");

		if (isOptimisticLock()) {
			t.addText( "if ("  + m_instance + ".getChecksum()==0)  return insert();");

			ContentTemplate ifT=t.createIf(" selectOnPk("
			+ m_instance
			+ ").hashCode()!="
			+ m_instance
			+ ".getChecksum()");
			ifT.addText("throw new Exception(\"Record has been changed by another user\");");
		}
			
		StringBuffer sb=new StringBuffer();
		sb.append("mPrep=mCon.prepareStatement(");
		
		sb.append("\"UPDATE ").append(m_tableModel.getTableName()).append(" SET ");
		boolean first=true;
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			if (cm.getPrimaryKey()==0) {
				if (!first) {
					sb.append(",");
				}
				sb.append(cm.getColumnName() + "=?");
				first=false;
			}

		}

		sb.append(" WHERE ");
		for (int i=0;i<pk.length;i++) {
			if (i!=0) sb.append(" AND ");
			sb.append(pk[i].getColumnName() + "= ? ");
		}
		sb.append("\");");
		
		t.addText(sb.toString());

		int seq = 0;
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			if (cm.getPrimaryKey() == 0) {
				seq++;
				t.addText(
					"mPrep.set"
						+ StringHelper.initCaps(
							Types2Java.getJavaDataType(
								cm.getDataType(),
								cm.getColumnSize(),
								cm.getDecimalDigits()),
							' ')
						+ "("
						+ seq
						+ ","
						+ m_instance
						+ ".get"
						+ StringHelper.objectName(cm.getColumnName())
						+ "());");
			}
		}
		
		if (!pk[0].getColumnName().equals("ROWID")) {
			for (int i = 0; i < pk.length ; i++) {	
				seq++;
				t.addText(
					"mPrep.set"
					+ StringHelper.initCaps(
						Types2Java.getJavaDataType(
							pk[i].getDataType(),
							pk[i].getColumnSize(),
							pk[i].getDecimalDigits()),
						' ')
					+ "("
					+ seq
					+ ","
				 	+ m_instance
					+ ".get"
					+ StringHelper.objectName(pk[i].getColumnName())
					+ "());");
			}
		} else {
			seq++;
			t.addText("mPrep.setString(" + seq + ","
					+ m_instance
					+ ".get"
					+ StringHelper.objectName(pk[0].getColumnName())
					+ "());");
		}
		t.addText("");
		t.addText("int result=mPrep.executeUpdate();");
		t.addText("mPrep.close();");
		if (isOptimisticLock()) {
			t.addText(m_instance
					+ ".setChecksum( selectOnPk("
					+ m_instance
					+ ").hashCode());");
		}
		t.addText("return result;");
		
	}

	/**
	 * writes the normal insert statement, possibly with the optimistic locking code
	 * @see #setOptimisticLock(boolean)
	 */
	private void writeInsert() {
		MethodTemplate t=classTemp.createMethod("insert","int",null);
		t.addException("Exception");
		t.addText(m_objName
				+ " "
				+ m_instance
				+ " = ("
				+ m_objName
				+ ") mObj;");
		
		SqlColumnModel[] pk=getPrimaryKey(m_tableModel);
		StringBuffer condition=new StringBuffer();
		if (pk!=null  && pk.length>0 && !pk[0].getColumnName().equals("ROWID")) {
			for (int i=0;i<pk.length;i++) {
				if (i>0) condition.append (" && ");
				condition.append(m_instance+".get"
						+ StringHelper.objectName(pk[i].getColumnName())
						+ "() =="
						+ getEmptyForType(pk[i].getDataType()));
			}
			t.createIf(condition.toString()).addText(
				"throw new Exception(\"Cannot insert record with empty primary key\");"
			);
		}
		t.addText("mPrep=mCon.prepareStatement(");

		StringBuffer sql=new StringBuffer();
		StringBuffer sb=new StringBuffer();
		sql.append("\"INSERT INTO ").append(m_tableModel.getTableName()).append(" (");
		
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			if (!cm.getColumnName().equals("ROWID")) {
				if (i != 0) {
					sql.append(",");
					sb.append(",");
				}
				sql.append(cm.getColumnName());
				sb.append("?");
			}
		}

		sql.append(")");
		sb.append(")");

		sql.append(" VALUES ( ").append(sb.toString());
		sql.append("\");");
		t.addText(sql.toString());
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			if (!cm.getColumnName().equals("ROWID")) {
				t.addText(
					"mPrep.set"
						+ StringHelper.initCaps(
							Types2Java.getJavaDataType(
								cm.getDataType(),
								cm.getColumnSize(),
								cm.getDecimalDigits()),
							' ')
						+ "("
						+ cm.getOrdinalPosition()
						+ ","
						+ m_instance
						+ ".get"
						+ StringHelper.objectName(cm.getColumnName())
						+ "());");
			}
		}

		t.addText("");
		t.addText("int result=mPrep.executeUpdate();");
		t.addText("mPrep.close();");
		if (isOptimisticLock()) {
			t.addText(m_instance
					+ ".setChecksum("
					+ m_instance
					+ ".hashCode());");
		}
		
		t.addText("return result;");
		
	}
	
	/**
	 * writes the normal delete statement
	 */
	private void writeDeleteOnPk() {
		SqlColumnModel[] pk = getPrimaryKey(m_tableModel);
		if (pk==null) {
			throw new RuntimeException("Table model has no primary key, pseudo primary key or ROWID. It is not possible to generate the DataControl");
		}

		MethodTemplate t=classTemp.createMethod("deleteOnPk","int",null);
		t.addException("Exception");
		t.addText(m_objName
				+ " "
				+ m_instance
				+ " = ("
				+ m_objName
				+ ") mObj;");
		t.addText("mPrep=mCon.prepareStatement(");

		StringBuffer sql=new StringBuffer();
		sql.append("\"DELETE FROM ").append(m_tableModel.getTableName());
		sql.append(" WHERE ");
		
		for (int i=0;i<pk.length;i++) {
			if (i!=0) sql.append(" AND ");
			sql.append(pk[i].getColumnName() + "= ? ");
		}
		sql.append("\");");
		t.addText(sql.toString());

		int seq = 0;
		for (int i = 0; i <pk.length; i++) {
			
			seq++;
			t.addText(
				"\t\t mPrep.set"
					+ StringHelper.initCaps(
						Types2Java.getJavaDataType(
							pk[i].getDataType(),
							pk[i].getColumnSize(),
							pk[i].getDecimalDigits()),
						' ')
					+ "("
					+ seq
					+ ","
					+ m_instance
					+ ".get"
					+ StringHelper.objectName(pk[i].getColumnName())
					+ "());");
			
		}

		t.addText("");
		t.addText("int result=mPrep.executeUpdate();");
		t.addText("mPrep.close();");
		t.addText("return result;");
		
	}

	/**
	 * writes a select on parent which takes one parameter an object which
	 * should represent a parent in a relation model.
	 * <p>While referential integratie is not recognized the method will generate
	 * an empty <code>null</code> result
	 *
	 */
	private void writeSelectOnParent() {
		MethodTemplate t=classTemp.createMethod(Modifier.PUBLIC,
			"selectOnParent","Vector",new String[][] {{"Object"},{"o"}});
		t.addException("Exception");
		t.addText("return null;");
		
	}
	
	/**
	 * write the select statement for selecting a single record on the primary key
	 *
	 */
	private void writeSelectOnPk() {
		SqlColumnModel[] pk=getPrimaryKey(m_tableModel);
		if (pk==null) {
			throw new RuntimeException("Table model has no primary key, pseudo primary key or ROWID. It is not possible to generate the DataControl");
		}

		MethodTemplate t=classTemp.createMethod(Modifier.PUBLIC,
			"selectOnPk","Object",new String[][] {{"Object"},{"o"}});
		t.addException("Exception");

		t.addText(m_objName
				+ " "
				+ m_instance
				+ " = ("
				+ m_objName
				+ ") o;");
		
		
		t.addText("mPrep=mCon.prepareStatement(");
		StringBuffer sql=new StringBuffer();
		sql.append("\"SELECT ");
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			sql.append(cm.getColumnName());
			if (i != m_tableModel.getColumns().size() - 1) {
				sql.append(",");
			}
		}
		sql.append(" FROM " + m_tableModel.getTableName());
		sql.append(" WHERE ");
		
		for (int i=0;i<pk.length;i++) {
			if (i!=0) sql.append(" AND ");
			sql.append(pk[i].getColumnName() + "= ? ");
		}
		sql.append("\");");
		t.addText(sql.toString());
		int seq = 0;
		for (int i = 0; i <pk.length; i++) {
			
			seq++;
			t.addText(
				"mPrep.set"
					+ StringHelper.initCaps(
						Types2Java.getJavaDataType(
							pk[i].getDataType(),
							pk[i].getColumnSize(),
							pk[i].getDecimalDigits()),
						' ')
					+ "("
					+ seq
					+ ","
					+ m_instance
					+ ".get"
					+ StringHelper.objectName(pk[i].getColumnName())
					+ "());");
			
		}
		t.addText("");
		t.addText("ResultSet rs=mPrep.executeQuery();");
		t.addText("Object out=null;");
		IfTemplate tIf= t.createIf("rs.next()");
		tIf.addText("out= getObject(rs);");
		
		t.addText("mPrep.close();");
		t.addText("return out;");
	}

	/**
	 * writes the select statement for selecting all records
	 *
	 */
	private void writeSelectAll() {
		MethodTemplate t=classTemp.createMethod(Modifier.PUBLIC,
			"selectAll","Vector",(String[])null);
		t.addException("Exception");
		SqlColumnModel pk = null;
		t.addText("Vector v=new Vector();");
		t.addText("mPrep=mCon.prepareStatement(");

		StringBuffer sql=new StringBuffer();
		sql.append("\"SELECT ");

		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			sql.append(cm.getColumnName());
			if (i != m_tableModel.getColumns().size() - 1) {
				sql.append(",");
			}

		}
		sql.append(" FROM " + m_tableModel.getTableName() + "\"");
		sql.append(");");
		t.addText(sql.toString());
		t.addText("");
		t.addText("ResultSet rs=mPrep.executeQuery();");
		t.createWhile("rs.next()").addText("v.add(getObject(rs));");
		t.addText("mPrep.close(); ");
		t.createIf("v.isEmpty()").addText("return null;");
		t.addText("return v;");
	}

	/**
	 * writes the select statement for selecting a group of records
	 * on the basis of an org.brains2b.sql.Condition
	 * @see org.brains2b.data.Condition
	 */
	private void writeSelectFromCondition() {
		classTemp.addImport("org.brains2b.data.Condition");
		classTemp.addImport("org.brains2b.sql.oracle.OracleCondition");
		MethodTemplate t=classTemp.createMethod(Modifier.PUBLIC,
			"selectFromCondition","Vector", new String[][] {{"Condition"},{"con"}});
		t.addException("Exception");
		
		t.addText("Vector v=new Vector();");
		StringBuffer sql=new StringBuffer();
		sql.append("String select= \"SELECT ");
		for (int i = 0; i < m_tableModel.getColumns().size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) m_tableModel.getColumns().get(i);
			sql.append(cm.getColumnName());
			if (i != m_tableModel.getColumns().size() - 1) {
				sql.append(",");
			}
		}
		sql.append(" FROM ").append(m_tableModel.getTableName()).append("\";");
		t.addText(sql.toString());
		t.addText("mPrep=((OracleCondition) con).getPreparedStatement(mCon,select);");
		t.addText("");
		t.addText("ResultSet rs=mPrep.executeQuery();");
		t.createWhile("rs.next()").addText("v.add(getObject(rs));");
		
		t.addText("mPrep.close(); ");
		t.createIf("v.isEmpty()").addText("return null;");
		
		t.addText("return v;");
		
	}

	/**
	 * get an array of SqlColumnModel which contains the columns making up the primary
	 * key in the proper sequence.
	 * @param model
	 * @return SqlColumModel[]
	 */
	private SqlColumnModel[] getPrimaryKey(SqlTableModel model) {
		Vector v=new Vector();
		for (int i = 0; i < model.getColumns().size(); i++) {
			SqlColumnModel column = (SqlColumnModel) model.getColumns().get(i);
			if (column.getPrimaryKey() != 0)
				v.add(column);
		}
		if (v.isEmpty()) return null;
		SqlColumnModel[] primarykeys=new SqlColumnModel[v.size()];
		v.copyInto(primarykeys);
		return primarykeys;
	}

	/**
	 * get the empty value for {CLASS or PRIMITIVE} resp. <code>null</code> 
	 * or <code>0</code>
	 * @param type
	 * @return String a representation of the empty value for a class or primitive
	 */
	private String getEmptyForType(int type) {
		switch (Types2Java.getJavaType(type)) {
			case Types2Java.CLASS :
				return "null";
			default :
				return "0";
		}
	}
    /**
     * get the classname used for generating this data control class
     * @see org.brains2b.sql.toclass.SqlClassWriter#getClassName(String)
     */
    protected String getClassName(String name) {
        m_objName = StringHelper.objectName(name);
		return m_objName + "DC";
		
    }

	/**
	 * check if this Data Control class should be generated as read only
	 * @return boolean, true if class should be generated read only, false otherwise
	 */
	protected boolean isReadOnly() {
		return m_readOnly;
	}

	/**
	 * set to true if this Data Control class is to be generated read only, false is the default for
	 * tables, true for Views.
	 * <p>This will not influence the code generated for a View
	 * @param b boolean, true if you want it to generate read only, false otherwise
	 */
	public void setReadOnly(boolean b) {
		m_readOnly = b;
	}

	/**
	 * check if this Data Control should contain the code for optimistic locking.
	 * @return boolean, true if class should generate optimistic locking code, false otherwise
	 * @see Sql2DataC#setOptimisticLock(boolean)
	 */
	protected boolean isOptimisticLock() {
		return m_optimisticLock;
	}

	/**
	 * sets this class to generate the required code for using optimistic locking.
	 * <p>Optimistic locking is done by setting a checksum value on a data class whenever
	 * a record is selected or updated. This value is the same as the calculated hashCode.
	 * <BR>Before a record is updated the record is refetched and the checksum is compared
	 * to the hashCode of the newly fetched record. Whenever there is a change made to the database
	 * values will no longer match and a 'Record has been changed by another user' Exception is thrown.
	 * <p>For this to work correctly the Data class used for this control class should contain a <code>setChecksum(int)</code>
	 * and <code>getChecksum()</code> method, something which is automatically generated if 
	 * <code>setOptimisticLock(<b>true</b>)</code> is used on Sql2File.
	 * @param b boolean, true if optimistic locking should be used, false otherwise
	 * @see Sql2File
	 */
	public void setOptimisticLock(boolean b) {
		m_optimisticLock = b;
	}

}
