/*  Sql2File.java
 *  Copyright (c) 2001, Dennis Groenendijk
 *	Copyright (c) 2002-2004, Brains2B.org
 *  
 *  Created on 21 mei 2001, 2:32
 *  Created by: Dennis Groenendijk
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 /*  History
 *  25-05-2001 Ratl 						Add default constructor
 *                  								Generate single name for Class
 *  19-12-2002 dennis@brains2b.nl	Added writeHashCode functie
 *  23-01-2002 dennis@brains2b.nl	Added setDate( java.util.Date)
 *  02-02-2003 dennis@brains2b.nl	Fixed hashCode problem for float and double
  * 08-03-2003 		"			" 			Improved displaying info an logging
 *  03-11-2003 dennis@brains2b.nl	Rewrote to use TypeWriter
 *  22-03-2004 dennis@brains2b.nl 	Added option for optimistic lock
 * 												Fixed problem with equals methods and mutiple keys
 * 												Added javadoc, cleaned up a bit
 */
package org.brains2b.sql.toclass;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.util.Date;
import java.util.Vector;

import org.brains2b.log.Logger;
import org.brains2b.sql.meta.SqlColumnModel;
import org.brains2b.sql.meta.SqlTableModel;
import org.brains2b.sql.meta.Types2Java;
import org.brains2b.typewriter.ClassTemplate;
import org.brains2b.typewriter.CommentTemplate;
import org.brains2b.typewriter.ContentTemplate;
import org.brains2b.typewriter.IfTemplate;
import org.brains2b.typewriter.MethodTemplate;
import org.brains2b.util.StringHelper;

/** This class writes an SQL Table which is supplied through an <code>SqlTableModel</code>
 * to a data class, with the necessary accessors
 * <p>This class is called from <code>Sql2ClassConsole</code> and some other implementations
 * which supply the necessary information. If you use this class just to generate files look at 
 * <code>Sql2ClassConsole</code> first. 
 * <p>The naming of the new class file will be equal to the table name, but with initial
 *  caps. example: <i>EMPLOYEES will become Employees</i>. Underscores will be treated
 * as an instance to initialize caps again. example: <i>EMP_SALARIES will become EmpSalaries</i>
 * <p>The class is written to the given location and packageName. The location should not include the
 * package directory itself. If the package directory does not exist it will be created. If the location
 * does not exist you will get a <code>FileNotFoundException</code>
 * <p>The option <a href="#setOptimisticLock><code>setOptimisticLock</code></a> sets extra code used in the DataControl class to 
 * implement optimistic locking for this data. If you leave this to false, no locking mechanism is provided 
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.40 [Mar 22, 2004]
 * @see org.brains2b.sql.toclass.Sql2ClassConsole
 * @see org.brains2b.sql.meta.SqlTableModel
 * @see org.brains2b.sql.toclass.Sql2DataC
 * 
 */
public class Sql2File extends SqlClassWriter {

	private String className;
	private ClassTemplate classTemp;
	private String[][] members;
	private Vector tableColumns;
	private boolean optimisticLock;
	
	/**
	 * Constructor
	 * @param location File, set the location to write the file to, this location should exist before run is called
	 * on this class.
	 * @param packageName String, the packageName for the class, names should be separated by dots. The packagename
	 * is both used to set the package name of the class as to create or use the directory relative to location to write the file to.
	 * @param tm org.brains2b.sql.meta.SqlTableModel, the table model to base this dataclass on
	 */
	public Sql2File(File location, String packageName, SqlTableModel tm) {
		super(location, packageName);
		setTableModel(tm);
		tableColumns = tm.getColumns();

	}

	/**
	 * Constructor
	 * <p>This constructor does not set the TableModel, this is however required
	 * for this class to be able to generate the file. You should set the tablemodel
	 * by calling <code>setTableModel(org.brains2b.sql.meta.SqlTableModel)</code>
	 * @param location File, set the location to write the file to, this location should exist before run is called
	 * on this class.
	 * @param packageName String, the packageName for the class, names should be separated by dots. The packagename
	 * is both used to set the package name of the class as to create or use the directory relative to location to write the file to.
	 * @see #setTableModel(SqlTableModel)
	 */
	public Sql2File(File location, String packageName) {
		super(location, packageName);
	}

	/**
	 * overwritten to retrieve the columns for local use
	 * @see org.brains2b.sql.toclass.SqlClassWriter#setTableModel(SqlTableModel)
	 */
	public void setTableModel(SqlTableModel tableModel) {
		super.setTableModel(tableModel);
		tableColumns = tableModel.getColumns();
	}

	/**
	 * run does the actual work in this class.
	 * <p>Output is written to the <code>org.brains2b.log.Logger</code>. Dependant
	 * on the setting there the code will show up on <code>System.out</code>. This is
	 * the case if you do nothing
	 * @see java.lang.Runnable#run()
	 * @see org.brains2b.task.DefaultTask#setFinished(boolean)
	 * @see org.brains2b.log.Logger
	 */
	public void run() {
		className = getClassName(m_tableModel.getTableName());
		classTemp = new ClassTemplate(m_packageName, className, null, null);

		makeMembers();
		writeHeaderComment();
		writeDefaultConstructor();
		writeConstructor();
		writeAccessors();
		if (m_tableModel.hasPrimaryKey()) {
			writeEquals();
		}
		writeHashCode();
		writeMembers();
		try {
			classTemp.writeClass(m_location.toString());
			Logger.println("data class saved as " + className + ".java");
			setResult(Sql2File.RESULT_OK);
		} catch (IOException ex) {
			Logger.printException(ex);
			setResult(RESULT_ERROR);
			addMessage(ex.getMessage());
		}

		setFinished(true);
	}

	/**
	 * get the file which has been created. 
	 * <p>The result of this file is only valid after <code>run()</code>has finished
	 * @see org.brains2b.sql.toclass.SqlClassWriter#getResultfile()
	 * @see #run()
	 */
	public File getResultfile() {
		return classTemp.getCompleteFile();
	}

	/**
	 * create the className from the tablemodel and set it to a member
	 */
	private void makeName() {
		className = getClassName(m_tableModel.getTableName());
	}

	/**
	 * creates an Array of members {their member names, local names and Java data types} from the <code>SqlColumnModel</code>s
	 * @see org.brains2b.sql.meta.SqlColumnModel
	 */
	private void makeMembers() {

		members = new String[tableColumns.size()][3];
		for (int i = 0; i < tableColumns.size(); i++) {
			SqlColumnModel scm = (SqlColumnModel) tableColumns.get(i);
			members[i][0] = StringHelper.memberName(scm.getColumnName());
			members[i][1] = StringHelper.objectName(scm.getColumnName());
			if (scm.getColumnName().equals("ROWID")) {
				members[i][2] = "String";

			} else {
				members[i][2] =
					Types2Java.getJavaDataType(
						scm.getDataType(),
						scm.getColumnSize(),
						scm.getDecimalDigits());
			}
		}
	}

	/**
	 * write a default comment header to the result file
	 */
	private void writeHeaderComment() {
		CommentTemplate comment = classTemp.createComment();
		comment.addText(Sql2ClassConsole.VERSIONID);
		comment.addText("Class " + className);
		comment.addText("Generated from table: " + m_tableModel.getTableName());
		comment.addText("Schema: " + m_tableModel.getTableSchema());
		//TODO get database name
		comment.addText("Database: ");
		comment.addText("");
		comment.addText("@since: " + new Date().toString());

	}

	/**
	 * write the default empty constructor
	 */
	private void writeDefaultConstructor() {
		classTemp.createConstructor((String[]) null);
	}

	/**
	 * write a constructor for this data class with all the data members (the columns in the table)
	 * as parameters
	 */
	private void writeConstructor() {

		String[][] params = new String[2][members.length];
		for (int i = 0; i < members.length; i++) {
			params[0][i] = members[i][2];
			params[1][i] = varName(members[i][1]);
		}
		MethodTemplate constructor = classTemp.createConstructor(params);
		for (int i = 0; i < members.length; i++) {
			constructor.addText(
				members[i][0] + " = " + varName(members[i][1]) + ";");
		}
	}
	
	/**
	 * write a getter and setter for each data member (column in the table)
	 * <p>For dates two setters are written: one to set a Timestamp or Date (dependant on SqlColumnModel.columnType)
	 * and one to set a java.util.Date
	 * @see org.brains2b.sql.meta.SqlColumnModel#getColumnType
	 * @see org.brains2b.sql.meta.Types2Java#getJavaDataType(int, int, int)
	 */
	private void writeAccessors() {
		if (isOptimisticLock()) {
			writeGetter("m_checksum", "Checksum", "int");
			writeSetter("m_checksum", "Checksum", "int");
		}

		for (int i = 0; i < members.length; i++) {
			//TODO Write getter and setter comments
			writeGetter(members[i][0], members[i][1], members[i][2]);
			writeSetter(members[i][0], members[i][1], members[i][2]);
			if (members[i][2].equals("Timestamp")) {
				classTemp.addImport("java.sql.Timestamp");
				writeDateSetter(
					"Timestamp",
					members[i][0],
					members[i][1],
					"java.util.Date");
			}
			if (members[i][2].equals("Date")) {
				classTemp.addImport("java.sql.Date");
				writeDateSetter(
					"Date",
					members[i][0],
					members[i][1],
					"java.util.Date");
			}

		}
	}

	/**
	 * write an equals method for this class based on the primary keys of the table.
	 * <p>This ensures an equals method based on primary keys only. Otherwise the 
	 * hashCode value is used as in the default equals method.
	 * <p>This method is only called when the table this class is based on contains
	 * primary keys
	 */
	private void writeEquals() {

		ContentTemplate equ =
			classTemp
				.createMethod(
					Modifier.PUBLIC,
					"equals",
					"boolean",
					new String[][] { { "Object" }, {
				"o" }
		});
		equ.addText("boolean result=false;");
		IfTemplate ifTemp = equ.createIf("o instanceof " + className);
		ifTemp.addText(className + " l = (" + className + ") o;");
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < tableColumns.size(); i++) {
			SqlColumnModel cm = (SqlColumnModel) tableColumns.get(i);
			if (cm.getPrimaryKey() != 0) {
				if (cm.getPrimaryKey() > 1) {
					sb.append(" && ");
				}
				if (Types2Java.getJavaType(cm.getDataType())
					== Types2Java.PRIMITIVE) {
					sb.append(
						" this."
							+ members[i][0]
							+ " == l.get"
							+ members[i][1]
							+ "() ");
				} else {
					sb.append(
						" this."
							+ members[i][0]
							+ ".equals(l.get"
							+ members[i][1]
							+ "()) ");
				}

			}
		}
		ifTemp.createIf(sb.toString()).addText("result = true;");
		equ.addText("return result;");

	}

	/** 
	 * write members
	 */
	private void writeMembers() {
		if (isOptimisticLock()) {
			classTemp.createMember("int", "m_checksum");
		}

		for (int i = 0; i < members.length; i++) {
			classTemp.createMember(members[i][2], members[i][0]);
		}
	}

	/**
	 * write a hashCode which is based on every members value (for primitives)
	 * and hashCodes (for objects) added on odd members or substracted on even
	 * members. 
	 */
	private void writeHashCode() {
		ContentTemplate hash =
			classTemp.createMethod(
				Modifier.PUBLIC,
				"hashCode",
				"int",
				(String[]) null);
		hash.addText("return ");
		for (int i = 0; i < members.length; i++) {
			String sign = "";
			if (i > 0)
				sign = (i % 2 == 0 ? "+" : "-");
			if (members[i][2].equals("int")) {
				hash.addText(sign + varName(members[i][0]));
			} else if (
				members[i][2].equals("float")
					|| members[i][2].equals("double")) {
				hash.addText(sign + "(int)" + varName(members[i][0]));

			} else {
				hash.addText(
					sign
						+ "("
						+ varName(members[i][0])
						+ "==null?0:"
						+ varName(members[i][0])
						+ ".hashCode())");
			}

		}
		hash.addText(";");
	}

	/**
	 * write an individual getter
	 * @param mName String , member name (m_)
	 * @param vName String, variable name, used to make the method name
	 * @param type String, representation of the Java type
	 */
	private void writeGetter(String mName, String vName, String type) {
		ContentTemplate get =
			classTemp.createMethod(
				Modifier.PUBLIC,
				"get" + vName,
				type,
				(String[]) null);
		get.addText("return " + mName + ";");

	}

	/**
	 * write an individual setter
	 * @param mName String , member name (m_)
	 * @param vName String, variable name, used as parameter and to make the methodname
	 * @param type String, representation of the Java type
	 */
	private void writeSetter(String mName, String vName, String type) {
		ContentTemplate set =
			classTemp
				.createMethod(
					Modifier.PUBLIC,
					"set" + vName,
					"void",
					new String[][] { { type }, {
				varName(vName)
				}
		});
		set.addText(mName + " = " + varName(vName) + ";");
	}

	/**
	 * write an individual setter for a java.util.Date
	 * @param sqltype String, completely qualified name for either <code>java.sql.Timestamp</code>
	 * or <code>java.sql.Date</code>
	 * @param mName String , member name (m_)
	 * @param vName String, variable name
	 * @param type String, representation java.util.Date
	 * @see java.sql.Date
	 * @see java.sql.Timestamp
	 */
	private void writeDateSetter(
		String sqltype,
		String mName,
		String vName,
		String type) {
		ContentTemplate set =
			classTemp
				.createMethod(
					Modifier.PUBLIC,
					"set" + vName,
					"void",
					new String[][] { { type }, {
				varName(vName)
				}
		});
		IfTemplate ifTemp = set.createIf(varName(vName) + "!=null");
		ifTemp.addText(
			mName
				+ "= new "
				+ sqltype
				+ "(("
				+ varName(vName)
				+ ").getTime());");
		ifTemp.createElse().addText(mName + " = null;");

	}

	/**
	 * get the class name for the data class
     * @see org.brains2b.sql.toclass.SqlClassWriter#getClassName(String)
     */
    protected String getClassName(String name) {
        return StringHelper.objectName(name);
    }	
	/**
	 * does this class use optimistic locking
	 * @return true if optimistic locking should be used, false otherwise
	 */
	protected boolean isOptimisticLock() {
		return optimisticLock;
	}

	/**
	 * set this to true if optimistic locking should be used.
	 * <p>by default optimistic locking is not used.
	 * @param b boolean
	 */
	public void setOptimisticLock(boolean b) {
		optimisticLock = b;
	}

}
