/*  SqlClassWriter.java
 *	Copyright (c) 2002-2004, Brains2B.org
 *
 *	Created on: May 19, 2002
 *	Created by: Dennis Groenendijk
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 /* @#History
 * [When]			[Who]						[What]
 * Mar 24, 2004	dennis@brains2b.nl	fixed bug in isFileSame for version checking
 * 													changed implementation of getFile, added javadoc
 */
package org.brains2b.sql.toclass;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import org.brains2b.log.Logger;
import org.brains2b.sql.meta.SqlTableModel;
import org.brains2b.task.DefaultTask;
import org.brains2b.typewriter.KeyWords;

/** Abstract class which supplies some general functions to classes generating Class files 
 *  from a database source.
 *  <LI>Checks locations and uses or creates package directories
 *  <LI>Checks if current file already exists and check if file needs to be regenerated
 * 
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.40 [23-03-2004]
 */
public abstract class SqlClassWriter extends DefaultTask {
	
	protected File m_location;
	protected String m_packageName;
	protected SqlTableModel m_tableModel;

   /**
    * Constructor
    *  @param location File, set the location to write the file to, this location should exist before run is called
	 * on this class. If this is null the current directory is used
	 * @param packageName String, the packageName for the class, names should be separated by dots. The packagename
	 * is both used to set the package name of the class as to create or use the directory relative to location to write the file to.
	 * If this is null no package name is used when creating this file.
    */
    protected SqlClassWriter(File location, String packageName) {
        super();
        m_location=location;
        m_packageName=packageName;
    }

	/**
	 * get the class name to be generated for the type of file. 
	 * <p>Implementation is left to the class extending this class
	 * @param name String, the table name
	 * @return String the name of the new class with proper capitalization
	 */
	protected abstract String getClassName(String name);

	/**
	 * test if the file you want to create already exists based on the tablename
	 * supplied.
	 * <p>This method will return true on any <code>IOException</code> except
	 * <code>FileNotFoundException</code> to make sure a valid file exists after the 
	 * generation has finished.
	 * @param tableName String, the tableName
	 * @return true if fle already exists and the content appears to be the same as 
	 * expected for the new file by calling <code>isFileSame</code>
	 * @see #isFileSame(File)
	 */
 	public boolean fileExists(String tableName) {
 		try {
 			File fl=getFile(m_location,m_packageName, getClassName(tableName));
 			return isFileSame(fl);
 		} catch (FileNotFoundException fnex) {
 			return false;
 		} catch (IOException iex) {
 			Logger.printException(iex);
 			return true;
 		}
 	}
 	
 	/**
 	 * sets the model for the table used for generating the file
 	 * @param tableModel SqlTableModel, the table model used
 	 */
 	public void setTableModel(SqlTableModel tableModel) {
 		m_tableModel=tableModel;
 	}

	/**
	 * get the File name on the basis of the supplied location, packageName and className
	 * @param location File, the location.
	 * @param packageName String, the new or existing packageName. If new will be created at the
	 * location. If location is null the packageName will be ignored
	 * @param className String, the name of the class, a result of getClassName()
	 * @return File, the complete file for the new file
	 * @throws IOException thrown if <code>createPackageDir</code> fails
	 * @see #createPackageDir(File, String)
	 */
	protected File getFile(File location, String packageName, String className) throws IOException {
		File fl = null;
		if ( location != null && packageName != null ) {
			
			fl = new File(
						createPackageDir(location,packageName),
						className+ ".java");
		} else if ( location != null ){
			fl = new File(location,className + ".java");
		} else {
			fl = new File(className + ".java");
		}
		return fl;
	}

	/**
	 * creates the package directories if they do not exist at a given location. 
	 * @param location File, the location where the package dirs should be or should be created
	 * @param name String, the package name separated by dots. If left to null the location is returned
	 * by itself.
	 * @return File, the complete directory including de package dirs
	 * @throws IOException, thrown if the supplied location does not exist or if the package dirs
	 * do not exist or could not be created
	 */
	private File createPackageDir(File location, String name) throws IOException {
		
		//If the location does not exist fail.
		if (!location.exists()) throw new FileNotFoundException("location " + location + " does not exist");
		if (name==null) return location;
		
		name=name.replace('.','/');
		File dir=new File(location,name);
		//package alreay exists as directory
		if (dir.exists()) return dir;
		if (dir.mkdirs() && dir.exists()) return dir;
		throw new IOException("could not create directories for package"); 
	}

	/** 
	 * check if the proposed file should be considered being the same as a current
	 * file in the same location with the same package. 
	 * <UL> Rules are:
	 * <LI>If there is no file which is the same at that location return false
	 * <LI> If the version id in the comment of an existing file is not the same as for this 
	 * version return false
	 * <LI>If the file contains a <code>@#History</code>-tag somewhere in the comments, before the
	 * actual class declaration the file is flagged as manually changed and this will return true so it won't
	 * be overwritten
	 * @param fl File, the file where are trying to establish already exists
	 * @return true if non of the above rules are met
	 * @see #fileExists(String)
	 */
    private boolean isFileSame(File fl) {
		boolean result=true;
		if (fl.exists()) {
			try {
				BufferedReader bfr=new BufferedReader(new FileReader(fl));
				String line=bfr.readLine();
				if (line!=null && line.startsWith("/*") && bfr.readLine().indexOf(Sql2ClassConsole.VERSIONID)==-1) {
					result=false;
				}
				while(line!=null) {
					if (line.indexOf("@#History")!=-1) {
						result=true;
						break;
					}
					if (line.startsWith("public class")) break;
					line=bfr.readLine();
				}
				bfr.close();			
			} catch (IOException ex) {
				Logger.printException(ex);
			}
		}else {
			result=false;
		}
		
		return result;
	}

	/**
	 * check if a varName is not a java keyword and can be used as a variable name
	 * otherwise prepend <code>x_</code> to it and use that as variable name
	 * <p>This method also makes sure at least the first letter is a lowercase letter
	 * @param str String, the value which needs to be used a variable Name
	 * @return String, the variable name with at least the first letter in lower case
	 * @see KeyWords
	 */
	protected String varName(String str) {
		String name=str.substring(0, 1).toLowerCase() + str.substring(1);
		if (KeyWords.isReserved(name)) {
			name="x_"+name;
		}
		return name;
	}

	/** abstract method for returning the generated file
	 * <p>implementation is left to the classes extending this class 
	 * @return File, the newly created file
	 */
	public abstract File getResultfile();
}
